import re
from collections import OrderedDict
from json import dumps

from django.db import models
from django.http import HttpResponse
from jsonfield import JSONField

# https://developer.mozilla.org/en-US/docs/Web/HTTP/Methods
METHODS = (
	('GET', 'GET'),          # The GET method requests a representation of the specified resource. Requests using GET should only retrieve data.
	('HEAD', 'HEAD'),        # The HEAD method asks for a response identical to that of a GET request, but without the response body.
	('POST', 'POST'),        # The POST method is used to submit an entity to the specified resource, often causing a change in state or side effects on the server.
	('PUT', 'PUT'),          # The PUT method replaces all current representations of the target resource with the request payload.
	('PATCH', 'PATCH'),      # The PATCH method is used to apply partial modifications to a resource.
	('DELETE', 'DELETE'),    # The DELETE method deletes the specified resource.
	('CONNECT', 'CONNECT'),  # The CONNECT method establishes a tunnel to the server identified by the target resource.
	('OPTIONS', 'OPTIONS'),  # The OPTIONS method is used to describe the communication options for the target resource.
	('TRACE', 'TRACE'),      # The TRACE method performs a message loop-back test along the path to the target resource.
)


TIMEOUT = 504

# https://www.restapitutorial.com/httpstatuscodes.html
STATUSES = (
	(200, '200 OK'),
	(201, '201 Created'),
	(204, '204 No content'),
	(301, '301 Moved permanently'),
	(304, '304 Not modified'),
	(307, '307 Temporary redirect'),
	(400, '400 Bad request'),
	(401, '401 Unauthorized'),
	(403, '403 Forbidden'),
	(404, '404 Not found'),
	(409, '409 Conflict'),
	(500, '500 Internal server error'),
	(501, '501 Not implemented'),
	(TIMEOUT, 'Timeout (no response)'),
)


class Endpoint(models.Model):
	# Request
	method = models.CharField(max_length=8, choices=METHODS, default=METHODS[0][0])
	path = models.CharField(max_length=128, blank=False, null=False, help_text='The relative path of the url as a regular expression, i.e. /post/\d/')
	parameters = JSONField(blank=True, help_text='Enter request parameters as a json map')
	
	# Other
	require_authentication = models.BooleanField(default=False, null=False)
	explanation = models.TextField(blank=True, help_text='Describe this endpoint for other administrators of django-mock-rest')
	
	@property
	def response_count(self):
		return self.responses.count()
	
	@property
	def path_pattern(self):
		# return self.path.pattern
		return self.path
	
	@property
	def path_regex(self):
		return re.compile('^{}$'.format(self.path))
	
	def __str__(self):
		return '{} {}'.format(self.method, self.path)
	
	def as_json(self):
		return OrderedDict((
			('method', self.get_method_display()),
			('path', self.path),
			('parameters', self.parameters),
			('require_authentication', self.require_authentication),
			('responses', tuple(resp.as_json() for resp in self.responses.all())),
		))


class Response(models.Model):
	endpoint = models.ForeignKey(Endpoint, related_name='responses', on_delete=models.CASCADE)
	weight = models.PositiveIntegerField(default=1, help_text='The relative likelihood of this response being sent (0 is disabled)')
	status = models.PositiveSmallIntegerField(choices=STATUSES, default=STATUSES[0][0])
	data = JSONField(blank=True, help_text='Enter a static json response')
	
	def __str__(self):
		return '{} ({})'.format(self.get_status_display(), self.weight)
	
	def is_timeout(self):
		return self.status == TIMEOUT
	
	def as_http_response(self):
		return HttpResponse(
			content=dumps(self.data, indent=4),
			status=self.status,
			content_type='application/json'
		)
	
	def as_json(self):
		return OrderedDict((
			('weight', self.weight),
			('status', self.get_status_display()),
			('data', self.data),
		))



# Generated by Django 2.1.1 on 2018-09-25 21:12

from django.db import migrations, models
import django.db.models.deletion
import jsonfield.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Endpoint',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('method', models.CharField(choices=[('GET', 'GET'), ('HEAD', 'HEAD'), ('POST', 'POST'), ('PUT', 'PUT'), ('PATCH', 'PATCH'), ('DELETE', 'DELETE'), ('CONNECT', 'CONNECT'), ('OPTIONS', 'OPTIONS'), ('TRACE', 'TRACE')], default='GET', max_length=8)),
                ('path', models.CharField(help_text='The relative path of the url as a regular expression, i.e. /post/\\d/', max_length=128)),
                ('parameters', jsonfield.fields.JSONField(blank=True, help_text='Enter request parameters as a json map')),
                ('require_authentication', models.BooleanField(default=False)),
                ('explanation', models.TextField(blank=True, help_text='Describe this endpoint for other administrators of django-mock-rest')),
            ],
        ),
        migrations.CreateModel(
            name='Response',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('weight', models.PositiveIntegerField(default=1)),
                ('status', models.PositiveSmallIntegerField(choices=[(200, '200 OK'), (201, '201 Created'), (204, '204 No content'), (301, '301 Moved permanently'), (304, '304 Not modified'), (307, '307 Temporary redirect'), (400, '400 Bad request'), (401, '401 Unauthorized'), (403, '403 Forbidden'), (404, '404 Not found'), (409, '409 Conflict'), (500, '500 Internal server error'), (501, '501 Not implemented'), (999, 'Timeout (no response)')], default=200)),
                ('data', jsonfield.fields.JSONField(blank=True, help_text='Enter a static json response')),
                ('endpoint', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='responses', to='django_mock_rest.Endpoint')),
            ],
        ),
    ]



from random import random


def match_paths(endpoints, path):
	found = []
	for endpoint in endpoints:
		print(endpoint.path_regex, path, bool(endpoint.path_regex.fullmatch(path)))
		if endpoint.path_regex.fullmatch(path) is not None:
			found.append(endpoints)
	return found

def choose_response(responses):
	total = sum(resp.weight for resp in responses)
	rem = random() * total
	for resp in responses:
		rem -= resp.weight
		if rem < 0:
			return resp
	return None


if __name__ == '__main__':
	"""
	Verification that the random results are kind of reasonable.
	"""
	class R:
		def __init__(self, w):
			self.weight = w
	q = (
		R(1),
		R(2),
		R(3),
	)
	r = {1: 0, 2: 0, 3: 0}
	for _ in range(100000):
		r[choose_response(q).weight] += 1
	from json import dumps
	print(dumps(r, indent=2))




from django.conf import settings


def is_enabled():
	enabled = getattr(settings, 'DJANGO_MOCK_REST', None)
	if enabled is True:
		return True
	if enabled is False:
		return False
	if enabled is None:
		return settings.DEBUG
	raise Exception("Incorrect configuration - DJANGO_MOCK_REST must be True, False or None")



from json import dumps
from time import sleep

from django.contrib.auth.decorators import login_required
from django.http import HttpResponse
from django.views.decorators.http import require_GET

from django_mock_rest.config import is_enabled
from django_mock_rest.models import Endpoint
from django_mock_rest.search import match_paths, choose_response


@require_GET
@login_required
def django_mock_rest_api_index(request):
	if not is_enabled():
		return HttpResponse("Mock API not enabled", status=500)
	endpoints = Endpoint.objects.prefetch_related()
	index = tuple(endpoint.as_json() for endpoint in endpoints)
	return HttpResponse(
		content=dumps(index, indent=4),
		content_type='application/json'
	)


def django_mock_rest_api(request, path):
	if not is_enabled():
		return HttpResponse("Mock API not enabled", status=500)
	path = "/" + path
	# Filter by method
	endpoints = Endpoint.objects.filter(method=request.method)
	# Filter by path, regex-based
	matches = match_paths(endpoints, path)
	if len(matches) > 1:
		return HttpResponse("Configuration error: more than one match for this method & path", status=500)
	if len(matches) == 0:
		return HttpResponse("No such mock endpoint: {} {}".format(request.method, path), status=404)
	endpoint = matches[0][0]
	# Find a response
	response = choose_response(endpoint.responses.all())
	if response is None:
		return HttpResponse("Configuration error: no response with weight >0 configured for this endpoint", status=500)
	if response.is_timeout():
		print('Timeout response; waiting 61s')
		sleep(61)
	return response.as_http_response()



from django.contrib import admin
from django import forms

from django_mock_rest.models import Endpoint, Response


class ResponseInline(admin.StackedInline):
	model = Response
	fields = ('weight', 'status', 'data',)
	extra = 1


class EndpointAdminForm(forms.ModelForm):
	def clean(self):
		path = self.cleaned_data['path']
		if not path.startswith('/'):
			path = '/' + path
		if not path.endswith('/'):
			path = path + '/'
		self.cleaned_data['path'] = path
		return self.cleaned_data


class EndpointAdmin(admin.ModelAdmin):
	model = Endpoint
	fields = ('method', 'path', 'explanation', 'require_authentication',)
	# fields = ('method', 'path', 'parameters', 'explanation', 'require_authentication',)
	list_display = ('method', 'path_pattern', 'response_count',)
	list_display_links = list_display
	list_filter = ('method',)
	form = EndpointAdminForm
	inlines = [
		ResponseInline,
	]


admin.site.register(Endpoint, EndpointAdmin)



from django.conf.urls import url

from django_mock_rest.views import django_mock_rest_api, django_mock_rest_api_index

urlpatterns = [
	url(r'^~$', django_mock_rest_api_index, name='django_mock_rest_index'),
	url(r'^(?P<path>[\w\-+/]*)$', django_mock_rest_api, name='django_mock_rest'),
]



# -*- coding: utf-8 -*-

from setuptools import setup

with open('README.rst', 'r') as fh:
	readme = fh.read()

setup(
	name='django-mock-rest',
	description='Simple way to create static mock data at rest api endpoints in the Django admin',
	long_description=readme,
	url='https://github.com/mverleg/django_mock_rest',
	author='Mark V',
	maintainer='(the author)',
	author_email='mverleg.noreply@gmail.com',
	license='Revised BSD License (LICENSE.txt)',
	keywords=['django', 'mock', 'test', 'unit-test', 'django-admin', 'rest', 'rest-api', 'api', 'http'],
	version='1.1',
	packages=['django_mock_rest', 'django_mock_rest.migrations',],
	include_package_data=True,
	zip_safe=True,
	classifiers=[
		'Development Status :: 5 - Production/Stable',
		'Intended Audience :: Developers',
		'Natural Language :: English',
		'License :: OSI Approved :: BSD License',
		'Operating System :: OS Independent',
		'Programming Language :: Python',
		'Programming Language :: Python :: 3',
		'Programming Language :: Python :: 3.4',
		'Programming Language :: Python :: 3.5',
		'Programming Language :: Python :: 3.6',
		'Programming Language :: Python :: 3.7',
		'Topic :: Software Development :: Libraries :: Python Modules',
	],
	install_requires= [
		'django>=2.0',
		'django-regex-field>=1.2',
		'jsonfield>=2.0',
	]
)
